import os
import scanpy as sc
import pandas as pd
from combat.pycombat import pycombat
import numpy as np
import warnings
import matplotlib.pyplot as plt
from sklearn.preprocessing import scale
from tqdm import tqdm
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from scipy.stats import spearmanr
import tensorflow as tf
from tensorflow.keras import layers
import random
import anndata as ad
def preprocess_and_plot_umap_metabolism(bulk_data, data, anno, metabolism_genes, save_path=None):
    """
    Preprocess gene expression data and generate UMAP images containing metabolic related genes

    Parameter:
    - bulk_data (DataFrame): DataFrame for batch data, with each column representing one sample
    - data (DataFrame): DataFrame for single cell data, where each row represents a gene and each column represents a cell
    - anno (DataFrame): DataFrame containing cell annotation information
    - metabolism_genes (DataFrame): DataFrame containing metabolic related genes, at least containing the 'gene name' column
    - save_path (str, optional): The path and file name for saving UMAP images. If not provided, it will be saved in the current path by default
    """
    print("Reading metabolic related genes")
    metabolism_genes = pd.read_csv(metabolism_genes)
    print("Successfully read")
    # Extracting metabolic related genes
    common_genes_metabolism = list(set(bulk_data.index).intersection(set(metabolism_genes['genename'])))
    bulk_metabolism = bulk_data.loc[common_genes_metabolism, :]
    data_metabolism = data.loc[common_genes_metabolism, :]
    num_common_genes_metabolism = len(common_genes_metabolism)
    print(f"Retained {num_common_genes_metabolism } metabolic related genes")
    # Create an AnaData object
    adata = sc.AnnData(X=data_metabolism.T)

    # Add cell annotation information to AnaData
    adata.obs['celltype'] = ''
    for idx, row in anno.iterrows():
        adata.obs.loc[row['cellname'], 'celltype'] = row['celltype']

    # pre-processing
    print("Start preprocessing data...")
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.highly_variable_genes(adata, n_top_genes=1500)
    sc.pp.scale(adata, max_value=10)
    sc.pp.pca(adata, n_comps=50, use_highly_variable=True)
    sc.pp.neighbors(adata)
    sc.tl.umap(adata)
    print("Preprocessing completed...")

    # Save UMAP image
    if save_path is None:
        save_path = os.path.join(os.getcwd(), "Orig_UMAP_Metabolism.png")
    else:
        save_path = os.path.join(os.getcwd(), save_path)

    print(f"Metabolism_ UMAP image will be saved to: {save_path}")
    sc.pl.umap(adata, color=['celltype'], legend_loc='on data', title='', frameon=True, show=False)
    plt.savefig(save_path)
    plt.close()
    layers_num = len(common_genes_metabolism)
    return bulk_metabolism,data_metabolism,common_genes_metabolism,layers_num


def prepare_train_val_data(data_metabolism, anno, batch_size):
    """
    Prepare training and validation set data, and generate batch data

    Parameter:
    - data_metabolism (DataFrame): DataFrame containing metabolic data, row name is cell name
    - anno (DataFrame): DataFrame containing cell annotation information needs to include the 'celltype' column
    - batch_size (int, optional): Batch size, default to 300

    Return:
    - train_data_batches (list): A list of training set batch data, each batch containing(batch, decoder_inputs)。
    - val_data_batches (list): A list of validation set batch data, each batch containing(batch, decoder_inputs)。
    """

    # Encoding cell types with unique heat
    enc = OneHotEncoder()
    cell_annotation_encoded = enc.fit_transform(anno[['celltype']])
    cell_annotation_encoded = pd.DataFrame(cell_annotation_encoded.toarray(),
                                           columns=enc.categories_[0])
    print('Divide the dataset into training and testing sets')
    # Assign row names to cells_ annotation_ encoded
    cell_annotation_encoded.index = data_metabolism.T.index

    # Merge metabolic data with cell type information encoded by unique heat
    data_metabolism_train = pd.concat([data_metabolism.T, cell_annotation_encoded], axis=1)

    # Divide training and validation sets
    train_data, val_data = train_test_split(data_metabolism_train, test_size=0.2, random_state=30)

    # Generate training set batch data
    train_data_batches = []
    for i in range(0, train_data.shape[0], batch_size):
        batch = train_data.iloc[i:i + batch_size].values
        decoder_inputs = batch[:, :-cell_annotation_encoded.shape[1]]
        train_data_batches.append((batch, decoder_inputs))

    # Generate validation set batch data
    val_data_batches = []
    for i in range(0, val_data.shape[0], batch_size):
        batch = val_data.iloc[i:i + batch_size].values
        decoder_inputs = batch[:, :-cell_annotation_encoded.shape[1]]
        val_data_batches.append((batch, decoder_inputs))
    return train_data_batches, val_data_batches

class CVAE(tf.keras.Model):
    def __init__(self, latent_dim, condition_dim,layers_num):
        super(CVAE, self).__init__()

        self.latent_dim = latent_dim
        self.condition_dim = condition_dim

        # encoder
        self.encoder = tf.keras.Sequential([
            layers.Input(shape=(layers_num,)),
            layers.Dense(1024, activation='relu'),
            layers.Dense(512, activation='relu'),
            layers.Dense(256, activation='relu'),
            layers.Dense(128, activation='relu')
        ])

        # decoder
        self.decoder = tf.keras.Sequential([
            layers.Input(shape=(latent_dim + condition_dim,)),
            layers.Dense(128, activation='relu'),
            layers.Dense(256, activation='relu'),
            layers.Dense(512, activation='relu'),
            layers.Dense(1024, activation='relu'),
            layers.Dense(layers_num, activation='sigmoid')
        ])

    # Reparameterization
    def sampling(self, z_mean, z_log_var):
        epsilon = tf.keras.backend.random_normal(shape=(tf.shape(z_mean)[0], self.latent_dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon

    def call(self, inputs):
        x = inputs[0]
        c = inputs[1]

        # Encoder output z_ Mean and z_ log_ var
        h = self.encoder(x)
        z_mean = layers.Dense(self.latent_dim)(h)
        z_log_var = layers.Dense(self.latent_dim)(h)

        # Reparameterization to obtain hidden vector z
        z = self.sampling(z_mean, z_log_var)

        # Splice hidden vectors and conditional vectors as decoder input
        z_cond = tf.concat([z, c], axis=1)

        # Decoder outputs reconstruction results
        recon_x = self.decoder(z_cond)

        return recon_x, z_mean, z_log_var


def get_condition_dim(cell_type_proportions_df):
    return len(cell_type_proportions_df.columns.tolist())


def train_cvae(train_data_batches, val_data_batches, latent_dim, condition_dim, epochs, learning_rate,layers_num):
    # Definition model
    cvae = CVAE(latent_dim, condition_dim,layers_num)
    # Define optimizer and loss function
    optimizer = tf.keras.optimizers.Adam(learning_rate)
    mse_loss_fn = tf.keras.losses.MeanSquaredError()

    # Training model
    train_loss_results = []
    val_loss_results = []
    best_val_loss = float("inf")
    best_model = None

    # Create progress bar
    progress_bar = tqdm(total=epochs, desc="Training Progress")

    for epoch in range(epochs):
        epoch_loss_avg = tf.keras.metrics.Mean()

        # At the beginning of each epoch, disrupt the order of training data
        np.random.shuffle(train_data_batches)

        for step, batch in enumerate(train_data_batches):
            x = batch[1]
            c = batch[0][:, -condition_dim:]

            # Calculate gradient and update parameters
            with tf.GradientTape() as tape:
                # Calculate reconstruction error
                recon_x, z_mean, z_log_var = cvae((x, c,layers_num))
                mse_loss = mse_loss_fn(x, recon_x)

                # Calculate KL divergence error
                kl_loss = -0.5 * tf.reduce_mean(1 + z_log_var - tf.square(z_mean) - tf.exp(z_log_var))
                total_loss = mse_loss + 0.25 * kl_loss

            gradients = tape.gradient(total_loss, cvae.trainable_variables)
            optimizer.apply_gradients(zip(gradients, cvae.trainable_variables))

            # Record the loss value for each step
            epoch_loss_avg.update_state(total_loss)

        train_loss_results.append(epoch_loss_avg.result().numpy())

        # Calculate loss values on the validation set
        val_loss_avg = tf.keras.metrics.Mean()
        for val_batch in val_data_batches:
            val_x = val_batch[1]
            val_c = val_batch[0][:, -condition_dim:]

            val_recon_x, val_z_mean, val_z_log_var = cvae((val_x, val_c,layers_num))
            val_mse_loss = mse_loss_fn(val_x, val_recon_x)
            val_kl_loss = -0.5 * tf.reduce_mean(1 + val_z_log_var - tf.square(val_z_mean) - tf.exp(val_z_log_var))
            val_total_loss = val_mse_loss + 0.25 * val_kl_loss

            val_loss_avg.update_state(val_total_loss)

        val_loss = val_loss_avg.result().numpy()
        val_loss_results.append(val_loss)
        if val_loss < best_val_loss:
            best_val_loss = val_loss
            best_model = cvae

        # Update progress bar
        progress_bar.set_postfix({"Train Loss": epoch_loss_avg.result().numpy(), "Val Loss": val_loss})
        progress_bar.update(1)

    progress_bar.close()

    print("Training finished!")
    print(f"Best validation loss: {best_val_loss:.4f}")
    print(f"Saved best model to best_model.h5")

    return train_loss_results, val_loss_results, best_model



import os
import scanpy as sc
import pandas as pd
from combat.pycombat import pycombat
import numpy as np
import warnings
import matplotlib.pyplot as plt
from sklearn.preprocessing import scale
from tqdm import tqdm
from scipy.optimize import nnls
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from scipy.stats import spearmanr
import tensorflow as tf
from tensorflow.keras import layers
import random
import anndata as ad
import argparse

from Dataloader import load_data,filter_genes,preprocess_and_plot_umap
from CVAEmodel import preprocess_and_plot_umap_metabolism,prepare_train_val_data,CVAE,get_condition_dim,train_cvae
from Deconvolution import scale_data,compute_cell_type_proportions
from GenerateData import simulate_bulk_data_pre,normalize_columns_to_sum_one,generate_synthetic_data,calculate_gene_expression_correlations
from Simulation import simulate_bulk_data,generate_umap_plots
from Std_choose import loop_choose_std,select_image


warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=UserWarning)
warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser(description='ScMet: A package used to generate sc RNA seq data that can represent Bulk RNA seq data, involving NNLS and CVAE')

parser.add_argument('--sc_data_dir', nargs='?',  type=str, default=None,
                    help='List of file names for storing single cell data.')
parser.add_argument('--sc_data_anno', nargs='?', type=str, default=None, metavar='<input_directory>',
                    help='List of cell annotation information file locations for storing single cell data pairing.')
parser.add_argument('--bulk_data', nargs='?', type=str, default=None, metavar='<data_directory>',
                    help='Bulk RNA seq data file location.')
parser.add_argument('--Meta_gene_dir', nargs='?', type=str, default= None,
                    help='Metabolism-related gene information file storage location.')
parser.add_argument('--learning_rate', nargs='?', type=float, default=0.0001,
                    help='Controlling the magnitude of parameter updates at each step.')
parser.add_argument('--epochs', nargs='?', type=int, default=50,
                    help='User defined EPOCH (training iteration).')
parser.add_argument('--batch_size', nargs='?', type=int, default=1440,
                    help='Number of cells per training.')
parser.add_argument('--Rm_batch_cell_number', nargs='?', type=int, default=2000,
                    help='Generate simulated Bulk RNA seq data and use it to remove the number of cells from the batch.')
parser.add_argument('--Rm_batch_sample_number', nargs='?', type=int, default=10,
                    help='Generate simulated Bulk RNA seq data and use it for sample size removal from batches.')
parser.add_argument('--Number_gene_deconvolution', nargs='?', type=int, default=250,
                    help='Number of genes retained by various cells for deconvolution.')
parser.add_argument('--Number_sm', nargs='?', type=int, default=3000,
                    help='Number of cells producing simulated small sample single-cell data.')
parser.add_argument('--Number_Nr', nargs='?', type=int, default=100000,
                    help='Number of cells producing simulated normal sample single-cell data.')
parser.add_argument('--latent_dim', nargs='?', type=int, default=64,
                    help='Refers to the dimensionality of the latent space in a machine learning model.')
parser.add_argument('--Std', nargs='?', type=int, default=1,
                    help='Standard deviation for single-cell data generation, multiple adjustments required.')
parser.add_argument('--num_simulations', nargs='?', type=int, default=20,
                    help='Number of times to fit Bulk RNA seq data using single cell data.')
parser.add_argument('--num_simulations_cell', nargs='?', type=int, default=5000,
                    help='The number of cells used to fit Bulk RNA seq data using single cell data.')

arg = parser.parse_args()

sc_data_dir = arg.sc_data_dir
if sc_data_dir is None:
    current_path = os.getcwd()
    parent_directory = os.path.dirname(current_path)
    resource_directory = os.path.join(parent_directory, 'Resource')
    sc_data_dir = os.path.join(resource_directory, 'CS1.csv')

sc_data_anno = arg.sc_data_anno
if sc_data_anno is None:
    current_path = os.getcwd()
    parent_directory = os.path.dirname(current_path)
    resource_directory = os.path.join(parent_directory, 'Resource')
    sc_data_anno = os.path.join(resource_directory, 'AN1.csv')

bulk_data_dir = arg.bulk_data
if bulk_data_dir is None:
    current_path = os.getcwd()
    parent_directory = os.path.dirname(current_path)
    resource_directory = os.path.join(parent_directory, 'Resource')
    bulk_data_dir = os.path.join(resource_directory, 'Bulk_1.csv')


Meta_gene = arg.Meta_gene_dir
if Meta_gene is None:
    current_path = os.getcwd()
    parent_directory = os.path.dirname(current_path)
    resource_directory = os.path.join(parent_directory, 'Resource')
    Meta_gene = os.path.join(resource_directory, 'metabolism_gene_total.csv')

Rm_batch_cell_number = arg.Rm_batch_cell_number
Rm_batch_sample_number = arg.Rm_batch_sample_number
Number_gene_deconvolution = arg.Number_gene_deconvolution
batch_size = arg.batch_size
epochs = arg.epochs
latent_dim = arg.latent_dim
learning_rate = arg.learning_rate
Number_sm = arg.Number_sm
Number_Nr = arg.Number_Nr
Std = arg.Std
num_simulations = arg.num_simulations
num_simulations_cell = arg.num_simulations_cell

print('Step 1: Import and preprocess the data')
sc_data,anno,bulk_data= load_data(sc_data_dir,sc_data_anno,bulk_data_dir)
bulk_data,sc_data= filter_genes(bulk_data,sc_data)
adata = preprocess_and_plot_umap(sc_data,anno)

print('Step 2: Solve for cell specific expression genes and deconvolute the proportion of cells')
scaled_matrix = scale_data(bulk_data)
cell_type_proportions_df = compute_cell_type_proportions(adata,scaled_matrix,Number_gene_deconvolution)

print('Step 3: Preserve metabolic related genes, generate training data, and train the CAVE model')
bulk_metabolism,data_metabolism,common_genes_metabolism,layers_num = preprocess_and_plot_umap_metabolism(bulk_data,sc_data,anno,Meta_gene)
train_data_batches,val_data_batches = prepare_train_val_data(data_metabolism,anno,batch_size)
batch_rm_data = simulate_bulk_data_pre(sc_data,bulk_data,Rm_batch_sample_number,Rm_batch_cell_number)
sc_normalized_data = normalize_columns_to_sum_one(batch_rm_data,common_genes_metabolism)
condition_dim = get_condition_dim(cell_type_proportions_df)
train_loss_results, val_loss_results, best_model = train_cvae(train_data_batches, val_data_batches, latent_dim, condition_dim, epochs, learning_rate,layers_num)

print('Step 4: Generate single cell data using the trained CVAE model')
print('Now we will generate small sample data for you based on different standard deviations. Please choose the existing standard deviation or input a new standard deviation based on the quality of the UMAP graph clustered by the generated data to achieve the best generation effect')
selected_index = Std
selected_type = 'std'
while selected_type is not None:
    img, z_std_list = loop_choose_std(best_model, cell_type_proportions_df, common_genes_metabolism, data_metabolism,
                                      anno, Number_sm, latent_dim, selected_index, None)
    selected_index, selected_type = select_image(img)
Std = z_std_list[selected_index]
generated_data, df = generate_synthetic_data(best_model, cell_type_proportions_df, common_genes_metabolism, data_metabolism, anno, Number_Nr, latent_dim, Std)

print('Step 5: Using generated data for fitting')
selected_gene_expression_df_total,selected_celltype,selected_samplename = simulate_bulk_data(cell_type_proportions_df,df, sc_normalized_data,common_genes_metabolism,num_simulations,num_simulations_cell)
adata= generate_umap_plots(selected_gene_expression_df_total, selected_celltype, selected_samplename)








import os
import scanpy as sc
import pandas as pd
from combat.pycombat import pycombat
import numpy as np
import warnings
import matplotlib.pyplot as plt
from sklearn.preprocessing import scale
from tqdm import tqdm
from scipy.optimize import nnls
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from scipy.stats import spearmanr
import tensorflow as tf
from tensorflow.keras import layers
import random
import anndata as ad

def simulate_bulk_data_pre(data, bulk_data, num_simulations=10, num_cells=600):
    bulk_sim_list = []
    for i in range(num_simulations):
        selected_cells = data.sample(n=num_cells, axis=1)
        bulk_sim = pd.DataFrame(np.sum(selected_cells, axis=1), columns=['Counts'])
        bulk_sim_list.append(bulk_sim)

    dataset_1 = bulk_data
    dataset_2 = pd.concat(bulk_sim_list,axis=1)
    df_expression = pd.concat([dataset_1,dataset_2],join="inner",axis=1)
    plt.boxplot(df_expression)
    plt.show()
    batch = []
    datasets = [dataset_1,dataset_2]
    for j in range(len(datasets)):
         batch.extend([j for _ in range(len(datasets[j].columns))])
    df_corrected = pycombat(df_expression,batch)
    # visualise results
    plt.boxplot(df_corrected)
    plt.show()
    sample=(dataset_1.shape[1])
    df_corrected_output = df_corrected.iloc[:,:sample]
    return df_corrected_output

def normalize_columns_to_sum_one(data,common_genes_metabolism):
    normalized_data = np.zeros_like(data)

    for col in range(data.shape[1]):
        column_sum = data.iloc[:, col].sum()
        normalized_data[:, col] = data.iloc[:, col] / column_sum

    normalized_data = pd.DataFrame(normalized_data)
    normalized_data.index = data.index
    normalized_data.columns = data.columns
    normalized_data = normalized_data.loc[common_genes_metabolism,:]
    return normalized_data


def generate_synthetic_data(best_model, cell_type_proportions_df, common_genes_metabolism, data_metabolism, anno,
                            num_samples=20000, latent_dim=64, z_std=2.5, save_path=None):
    """
    Generate synthetic data samples.

    Parameter:
    - best_model (CVAE): A trained CVAE model
    - cell_type_proportions_df (DataFrame): The proportion of each cell type in each sample, represented by the DataFrame with the sample name as the row and the cell type as the column.
    - common_genes_metabolism (list): List of Common Metabolic Genes
    - metabolism_df (DataFrame): DataFrame representing metabolic genes
    - num_samples (int, optional): Generate sample quantity, default to 20000
    - latent_dim (int, optional): The dimension of latent variables, default to 64

    Return:
    - generated_data (DataFrame): Generated synthetic data samples, including metabolic gene information and cell type labels
    """

    # Randomly generated latent variables and conditions
    z_mean = np.zeros(latent_dim)
    z = np.random.normal(loc=z_mean, scale=z_std, size=(num_samples, latent_dim))

    # Calculate the number of each cell type
    cell_type_counts = anno.groupby('celltype').size()
    cell_type_proportions_sc = cell_type_counts / cell_type_counts.sum()
    cell_type_proportions_df_sc = cell_type_proportions_sc.to_frame(name='proportions').reset_index()
    cell_type_proportions_df_sc = cell_type_proportions_df_sc.set_index('celltype')
    cell_type_proportions_df_transposed_sc = cell_type_proportions_df_sc.T
    cell_type_proportions_df_transposed_sc.reset_index(drop=True, inplace=True)
    cell_types_list_sc = cell_type_proportions_df_transposed_sc.columns
    cell_types = np.random.choice(cell_types_list_sc, size=num_samples, p=cell_type_proportions_sc.values)

    # Encoding cell types with OneHotEncoder
    encoder = OneHotEncoder(sparse=False)
    c = encoder.fit_transform(cell_types.reshape(-1, 1))

    # Merge latent variables and conditions into a tensor
    z_c = np.hstack([z, c])

    # Input CVAE model to generate sample data and cell type annotations
    x_pred, y_pred = best_model.decoder.predict(z_c), np.argmax(c, axis=1)
    print(f"Cell generation completed, total generation: {num_samples} cells")

    # Save the generated data and labels as DataFrame
    df = pd.DataFrame(x_pred)
    df['celltype'] = y_pred
    labels = cell_type_proportions_df.columns.tolist()
    label_dict = {i: label for i, label in enumerate(labels)}
    df['celltype'] = df['celltype'].replace(label_dict)
    generated_data = pd.DataFrame(x_pred, columns=common_genes_metabolism)
    print("Continue processing the generated data")
    # Process the generated data
    value_proportions = data_metabolism.T.apply(lambda x: x.value_counts(normalize=True))

    for col in generated_data.columns:
        # Filter out data with a ratio of 0 and sort in descending order
        proportions = value_proportions[col][value_proportions[col] > 0.0001].sort_index(ascending=False)
        values = proportions.index.values
        cum_proportions = np.cumsum(proportions.values)

        # Calculate the number of elements that should be assigned to each value based on proportion
        num_elements_per_value = (cum_proportions * generated_data.shape[0]).astype(int)
        num_elements_diff = np.diff(np.concatenate(([0], num_elements_per_value)))

        # Correct the number of the last element to match the number
        num_elements_diff[-1] = generated_data.shape[0] - np.sum(num_elements_diff[:-1])

        # Generate a list of all values based on the number of elements and arrange them in descending order
        all_values = np.repeat(values, num_elements_diff)

        # For generated_ Sort the current column of data in descending order and obtain the sorted index
        sorted_indices = np.argsort(-generated_data.loc[:, col])

        # Assign the arranged values to generated_ Data, keep row order unchanged
        generated_data.loc[sorted_indices, col] = all_values
    print(
        "Processing completed, conducting simulation data quality analysis. Please try different standard deviation values based on the UMAP clustering diagram")
    adata = ad.AnnData(X=generated_data, obs={'cell_type': df['celltype'].values})
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.highly_variable_genes(adata, n_top_genes=1000)
    sc.pp.scale(adata)
    sc.tl.pca(adata, svd_solver='arpack')
    sc.pp.neighbors(adata, n_neighbors=15, n_pcs=50)
    sc.tl.umap(adata)
    # Save UMAP image
    if save_path is None:
        save_path = os.path.join(os.getcwd(), "Simulation_UMAP_Metabolism_std_" + str(z_std) + ".png")
    else:
        save_path = os.path.join(os.getcwd(), save_path)

    print(f"Total generated: {num_samples} cells")
    print(f"Metabolism_ UMAP image will be saved to: {save_path}")
    print(f"The current standard deviation value is: {z_std},Please adjust according to the UMAP clustering effect")

    sc.pl.umap(adata, color=['cell_type'], legend_loc='on data', title='', frameon=True, show=False)
    plt.savefig(save_path)
    plt.close()
    sc.pl.umap(adata, color=['cell_type'], legend_loc='on data', title='', frameon=True)
    generated_data = pd.DataFrame(x_pred, columns=common_genes_metabolism)
    value_proportions = data_metabolism.T.apply(lambda x: x.value_counts(normalize=True))

    for col in generated_data.columns:
        # Filter out data with a ratio of 0 and sort in descending order
        proportions = value_proportions[col][value_proportions[col] > 0.0001].sort_index(ascending=False)
        values = proportions.index.values
        cum_proportions = np.cumsum(proportions.values)

        # Calculate the number of elements that should be assigned to each value based on proportion
        num_elements_per_value = (cum_proportions * generated_data.shape[0]).astype(int)
        num_elements_diff = np.diff(np.concatenate(([0], num_elements_per_value)))

        # Correct the number of the last element to match the number
        num_elements_diff[-1] = generated_data.shape[0] - np.sum(num_elements_diff[:-1])

        # Generate a list of all values based on the number of elements and arrange them in descending order
        all_values = np.repeat(values, num_elements_diff)

        # Sort the current column of generated_data in descending order and obtain the sorted index
        sorted_indices = np.argsort(-generated_data.loc[:, col])

        # Assign the arranged values to XX, keeping the row order unchanged
        generated_data.loc[sorted_indices, col] = all_values

    return generated_data, df

def calculate_gene_expression_correlations(data_metabolism, common_genes_metabolism,anno,generated_data,generated_labels,cell_type_proportions_df):
    # Convert raw data and generated data to DataFrame
    adata_raw_data = pd.DataFrame(data_metabolism.T, columns=common_genes_metabolism)
    adata_raw_cell_types = pd.DataFrame(anno, columns=['celltype'])
    adata_raw_cell_types.index = adata_raw_data.index

    generated_data_df = pd.DataFrame(generated_data, columns=common_genes_metabolism)
    generated_cell_types_df = pd.DataFrame(generated_labels, columns=['celltype'])

    # Merge raw data and generate data
    adata_raw = pd.concat([adata_raw_data, adata_raw_cell_types], axis=1)
    adata_generated = pd.concat([generated_data_df, generated_cell_types_df], axis=1)

    # Calculate the correlation of gene expression profiles for various cell types
    cell_types = cell_type_proportions_df.columns
    correlations = {}

    for cell_type in cell_types:
        # Extract raw data and generate data for specific cell types
        raw_cell_type_data = adata_raw[adata_raw['celltype'] == cell_type].drop('celltype', axis=1)
        generated_cell_type_data = adata_generated[adata_generated['celltype'] == cell_type].drop('celltype', axis=1)

        # Calculate the average of gene expression profiles between raw and generated data
        raw_mean = raw_cell_type_data.mean(axis=0)
        generated_mean = generated_cell_type_data.mean(axis=0)

        # Calculate Spearman correlation coefficient
        corr, _ = spearmanr(raw_mean, generated_mean)

        correlations[cell_type] = corr

    return correlations




from setuptools import setup

def readme_file():
    with open("README.rst") as rf:
        return rf.read()

setup(
    name='scMet',
    version='1.0.3',
    author='Gw yang',
    author_email='stu_gwyang@163.com',
    description='Fits bulk RNA-seq data using single-cell RNA-seq (scRNA-seq) data, enabling accurate estimation and analysis of gene expression profiles and metabolic flux balance',
    long_description='',
    url='https://github.com/your_username/your_package',
    packages=['scMet'],
    install_requires=[
        'scanpy',
        'pandas',
        'combat==0.2.1',
        'numpy',
        'matplotlib',
        'sklearn',
        'tqdm',
        'scipy',
        'tensorflow',
        'anndata'
    ],
    python_requires='>=3.8',
)
